import numpy as np
from sklearn.cluster import DBSCAN


class DbscanLoop:
    def __init__(self, eps, min_samples):
        """
        The implementation of the *noise free* implementation of DBscan, described in the paper.
        """
        self.eps = eps
        self.min_samples = int(min_samples)

    def fit(self, XS):
        XS_full = XS
        n_old = len(XS)
        idx_core = np.arange(len(XS))
        DB = DBSCAN(eps=self.eps, min_samples=self.min_samples).fit(XS)
        XS = XS[DB.core_sample_indices_]

        while (n_old != len(XS)) and (len(XS) > 0):
            n_old = len(XS)
            idx_core = idx_core[DB.core_sample_indices_]
            DB = DBSCAN(eps=self.eps, min_samples=self.min_samples).fit(XS)
            XS = XS[DB.core_sample_indices_]

        # all labels are noise
        labels_ = -1 * np.ones((len(XS_full),), dtype=int)
        # change the found labels to their correct value
        labels_[idx_core] = DB.labels_

        self.labels_ = labels_
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 30 19:18:32 2021

@author: andreas
"""
import numpy as np
import pandas as pd

"""
JUST SOME UTILS
"""


def GetLineOfOptima(df, x_selector, y_selector, includeAllColumns=False):
    """
    For each unique value in `x_selector`, find the entry which maximizes 'y_selector" and return a pd.DataFrame
    containing the indexes with the best entry, the unique 'x_selector' values and the associated 'y_selector' values.

    This gives the *line of optima*, i.e., the optimal entry for each unique value of `x_selector`.



    Parameters
    ----------
    df: pd.DataFrame
        the dataframe with the data.
    x_selector: str
        the name of the  column you want to explore (usually, in FINDER, is `threshold`);
    y_selector: str
        the name of the column you want to use as metric (in FINDER, it is usually 'similarityscore');
    includeAllColumns: bool, default = False
        if False, just return the values of `y_selector` for each point. If True, return all columns.
    Returns
    -------
    df_opt = pd.DataFrame
        a DataFrame with columns ['idx', 'x_selector', 'y_selector'].
        If includeAllColumns== True, contains all the columns of `df` instead.
    """

    # unique entries of 'x_selector'
    x_sel_sort = np.sort(np.unique(df[x_selector]))
    idxs = []

    # for each `x_sel`, find the index leading to the maximum `y_selector` value.
    for x_sel in x_sel_sort:
        mark_ = df[x_selector] == x_sel
        if np.sum(mark_) == 0:
            continue
        else:
            idxs.append((df.loc[mark_, y_selector]).idxmax())

    df_opt = pd.DataFrame()
    df_opt["idx"] = idxs

    if includeAllColumns:
        cols = df.columns
    else:
        cols = [x_selector, y_selector]

    # generate the dataframe with the optima only
    for c in cols:
        df_opt[c] = np.asarray(df.loc[idxs, c])

    return df_opt


def GetClusterDistribution(labels):
    """
    Return an array containing the number of points with a given label.

    Position `i` in the array account for label 'i`. Its value will be the number of elements with label `i`.
    NOTE: the points that are considered noise (label == -1) are not included.

    Parameters
    ----------
    labels: the labels

    Returns
    -------

    """
    cl_sizes = []
    for c in np.unique(labels):
        if c == -1:
            continue
        cl_sizes.append(np.sum(labels == c))
    return cl_sizes


# Print iterations progress
def printProgressBar(
        iteration,
        total,
        prefix="",
        suffix="",
        decimals=1,
        length=100,
        fill="█",
        printEnd="\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(
        100 * (iteration / float(total))
    )
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + "-" * (length - filledLength)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=printEnd)
    # Print New Line on Complete
    if iteration == total:
        print()



import time
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from sklearn.cluster import DBSCAN
from sklearn.neighbors import NearestNeighbors

from .AuxiliaryFunctions import GetLineOfOptima, printProgressBar
from .DbscanLoop import DbscanLoop

from .SimilarityScore import (
    getClusterSizesAll,
    getSimilarityScore,
    getSimilarityScoreByThreshold,
)


class Finder:
    def __init__(
            self,
            threshold: int = 10,
            points_per_dimension=15,
            algo="DbscanLoop",
            minmax_threshold=[5, 21],
            one_two_d="twoD",
            similarity_score_computation="threshold",
            log_thresholds=False,
            log_sigmas=True,
            adaptive_sigma_boundaries=False,
            decay=0.5,
            n_dim: int = None
    ):

        """
        The model class for the `FINDER` algorithm. It uses the sklearn API.
        It explores the two parameters (`threshold` and `sigma`) of the given `algo` and finds the best combination.
        Note that in the paper (and sometimes in the code) the parameters are
        called 'minPts'(=threshold) and 'epsilon'(=sigma) instead.

        Parameters
        ----------
        threshold: int, default = 10
            The default value for the `threshold` value, also called `minPts`.
        points_per_dimension:
            the number of values for each of the two axis  (`threshold` and `sigma`) of the phase space.
        algo: str, default="DbscanLoop"
            the algorithm to be used. Can be either "dbscan" or "Dbscanloop" todo: fix the capital D
        minmax_threshold: list, default = [3,30],
            The minimum and maximum values for the "threshold" parameter.
        one_two_d: str, default = "twoD"
            A string to define the kind of optimization that one wish to carry.
            Possible values are ["oneD", "oneD_thresholds", "twoD"]
        similarity_score_computation: str, default = 'total'
            Select the way in which the similarity score is computed:

            * `total`: compare each parameter combination with any other.
                The number of comparison will then be (p**2)*(p**2 - 1)/2 = O(p**4),
                where p is `points_per_dimension`.
            * 'threshold: compute the similarity only between parameter-combination sharing the same threshold
                (i.e., line by line).
                The number of comparison will then be p*(p)*(p- 1)/2 = O(p**3),
            Applies to the the "twoD" setting only
        log_thresholds: Bool, default = 'False'
            If `True`, the threshold-values are selected in log scale.
        log_sigmas: Bool, default = 'True'
            If `True`, the sigma-values are selected in log scale.
        decay: float, default = 0.5
            The value used to select the selected parameters.
        n_dim: int, default = None
            The dimensionality of your data, presumably 2D or 3D. If `None`, it will be inferred using the data.

        """

        self.threshold = threshold
        self.no_points_sigma = points_per_dimension
        self.no_points_thresholds = points_per_dimension
        self.algo = algo
        self.minmax_threshold = minmax_threshold
        assert one_two_d in [
            "oneD",
            "oneD_thresholds",
            "twoD",
        ], 'Possible values are ["oneD", "oneD_thresholds", "twoD"]'
        self.one_two_d = one_two_d
        self.decay = decay
        assert similarity_score_computation in [
            "total",
            "threshold",
        ], 'possible values are ["total", "threshold"]'
        self.similarity_score_computation = similarity_score_computation
        self.log_thresholds = log_thresholds
        self.log_sigmas = log_sigmas
        self.adaptive_sigma_boundaries = adaptive_sigma_boundaries
        self.n_dim = n_dim

    def fit(self, XC, XC_params=None, **kwargs):
        """
        Generate the labels using FINDER algorithm.

        **Steps**:
         1. Set Boundaries: Get min max of threshold and sigma
         2. Clustering: Compute phase spaces
         3. Postprocessing: Compute similarity score (can be skipped)

        Parameters
        ----------
        XC:
            Points to be clustered
        XC_params:
            The points used to compute the boundaries. If `Null`, `XC` will be used.
        kwargs

        Returns
        -------

        clustering labels

        """
        print("Analysing " + str(len(XC)) + " points")

        if XC_params is None:
            XC_params = XC

        if self.n_dim is None:
            self.n_dim = np.array(XC).shape[1]

        t_1 = time.time()

        # Step 1: Get min max of threshold and sigma
        if self.one_two_d == "oneD":
            params = self.__getParams_Sigmas(XC_params)
        if self.one_two_d == "oneD_thresholds":
            params = self.__getParams_Thresholds(XC_params)
        elif self.one_two_d == "twoD":
            params = self.__getParams_SigmasThresholds(XC_params)

        t_2 = time.time()

        # Step 2: Compute phase spaces
        phasespace = self.__phaseSpace(XC, params)
        self.phasespace = phasespace
        t_3 = time.time()

        # Step 3: Compute similarity score
        if ("skipSimilarityScore" in kwargs.keys()) and (
                kwargs["skipSimilarityScore"] == True
        ):
            print("skipping the similarity-score computation")
            data = self.__phaseSpacePostProcess(
                XC, phasespace, skipSimilarityScore=True
            )

            selth = self.phasespace.loc[
                np.argmin(
                    np.abs(self.phasespace["threshold"] - self.threshold)
                ),
                "threshold",
            ]
            mark = self.phasespace["threshold"] == selth
            selected_parameters = self.phasespace.loc[
                self.phasespace.loc[mark, "no_clusters"].idxmax()
            ]
            labels = selected_parameters["labels"]

        else:
            data = self.__phaseSpacePostProcess(XC, phasespace)
            # Step 3: Get parameterset
            if self.one_two_d == "twoD":
                labels, selected_parameters = self.__get_consensus_clustering(
                    data, XC
                )
            else:
                print("USING 1D OPTIMIZER")
                (
                    labels,
                    selected_parameters,
                ) = self.__get_consensus_clustering_1d(data, XC)

        t_4 = time.time()

        # labels,selected_parameters = self.__get_consensus_clustering2(data,XC);
        print(
            "Comp time Step 1 (set boundaries): "
            + str(np.round(t_2 - t_1, 2))
            + " seconds"
        )
        print(
            "Comp time Step 2 (clustering): "
            + str(np.round(t_3 - t_2, 2))
            + " seconds"
        )
        print(
            "Comp time Step 3 (postprocessing): "
            + str(np.round(t_4 - t_3, 2))
            + " seconds"
        )
        print("Selected parameters: " + str(selected_parameters))

        # Save data
        self.computationTimes = {
            "Step1": t_2 - t_1,
            "Step2": t_3 - t_2,
            "Step3": t_4 - t_3,
        }
        self.data = data
        self.labels = labels
        self.selected_parameters = selected_parameters

        return labels

    def ComputeClusters(self, sigma, threshold, XC):
        """

        Cluster the point using the selected `self.algo` (DBSCAN or Dbscanloop).
        This is basically a iteration of `self.algo` in a given phase-space point.

        Parameters
        ----------
        sigma:
            radius.
        threshold:
            number of point inside the radius.
        XC:
            data-points.

        Returns
        -------

        The labels

        """
        threshold = int(threshold)

        if self.algo == "dbscan":
            DB = DBSCAN(eps=sigma, min_samples=threshold).fit(XC)
            labels_ = DB.labels_

        elif self.algo == "DbscanLoop":
            DBL = DbscanLoop(eps=sigma, min_samples=threshold).fit(XC)
            labels_ = DBL.labels_

        else:
            print("ALGORITHM NOT RECOGNIZED !!")

        return labels_

    def __getParams_SigmasThresholds(self, XC):

        """
        Compute the parameter range for both `threshold` and `sigma`.

        Parameters
        ----------
        XC: the datapoints

        Returns
        -------
        pd.Dataframe
        """

        if self.adaptive_sigma_boundaries:
            minmax_sigma = self.__determine_sigma_boundaries_adaptive(XC)
        else:
            minmax_sigma = self.__determine_sigma_boundaries(XC)

        if self.log_sigmas:
            sigmas = self.__getLogDistribution(
                minmax_sigma[0], minmax_sigma[1], self.no_points_sigma
            )
        else:
            sigmas = np.linspace(
                minmax_sigma[0], minmax_sigma[1], self.no_points_sigma
            )

        if self.log_thresholds:
            thresholds = np.linspace(
                self.minmax_threshold[0],
                self.minmax_threshold[1],
                self.no_points_sigma,
            )
            thresholds = np.unique(np.round(thresholds))
        else:
            thresholds = np.arange(
                self.minmax_threshold[0], self.minmax_threshold[1]
            )

        # thresholds = self.__getLogDistribution(self.minmax_threshold[0],
        # self.minmax_threshold[1],
        # self.no_points_thresholds)

        # #LINEAR
        # #sigmas = np.linspace(minmax_sigma[0], minmax_sigma[1], self.no_points_sigma)
        # #thresholds = np.linspace(self.minmax_threshold[0], self.minmax_threshold[1], self.no_points_sigma)
        #
        # #thresholds = np.unique(np.round(thresholds))
        #
        # thresholds = np.arange(self.minmax_threshold[0], self.minmax_threshold[1])
        # print("done")

        s_all = []
        t_all = []
        for s in sigmas:
            for t in thresholds:
                s_all.append(s)
                t_all.append(t)

        print("Sigmas are:")
        print(sigmas)
        print("Thresholds are:")
        print(thresholds)

        params = pd.DataFrame(
            data={"sigma": np.asarray(s_all), "threshold": np.asarray(t_all)}
        )
        return params

    def __getParams_Sigmas(self, XC):

        """
        Compute the parameter range for `sigma` for a fixed `threshold` value.
        The value for `threshold` is 'self.threshold'.

        Parameters
        ----------
        XC: the datapoints

        Returns
        -------
        params:pd.Dataframe
            the values of the parameters that FINDER will explore.
        """

        if self.adaptive_sigma_boundaries:
            minmax_sigma = self.__determine_sigma_boundaries_adaptive(XC)
        else:
            minmax_sigma = self.__determine_sigma_boundaries(XC)

        if self.log_sigmas:
            sigmas = self.__getLogDistribution(
                minmax_sigma[0], minmax_sigma[1], self.no_points_sigma
            )
        else:
            sigmas = np.linspace(
                minmax_sigma[0], minmax_sigma[1], self.no_points_sigma
            )

        params = pd.DataFrame(
            data={
                "sigma": sigmas,
                "threshold": self.threshold * np.ones_like(sigmas),
            }
        )
        return params

    def __getParams_Thresholds(self, XC):

        """
        Compute the parameter range for `threshold` for a fixed `sigma` value.
        The value for `sigma` is computed using a KNN.

        Parameters
        ----------
        XC: the datapoints

        Returns
        -------
        params: pd.Dataframe
            the values of the parameters that FINDER will explore.
        """

        k = 10
        # initialize model
        neigh = NearestNeighbors(n_neighbors=k, n_jobs=-1)
        # train for getting nearest neighbour
        neigh.fit(XC)
        dist_, ind = neigh.kneighbors(XC)

        # We have no use of indices here
        # dist is a 2 dimensional array of shape (10000, 9) in which each row is list of length 9.
        # This row contain distances to all 9 nearest points. But we need distance to only 9th nearest point. So
        nPt_distance = [dist_[i][k - 1] for i in range(len(dist_))]

        # CD_sorted = np.sort(dist.squareform(dist.pdist(XC)),axis=1);
        sigma = np.quantile(nPt_distance, 0.5)

        if self.log_thresholds:
            thresholds = np.linspace(
                self.minmax_threshold[0],
                self.minmax_threshold[1],
                self.no_points_sigma,
            )
            thresholds = np.unique(np.round(thresholds))
        else:
            thresholds = np.arange(
                self.minmax_threshold[0], self.minmax_threshold[1]
            )

        params = pd.DataFrame(
            data={
                "sigma": sigma * np.ones_like(thresholds),
                "threshold": thresholds,
            }
        )
        return params

    def __determine_sigma_boundaries_adaptive(self, XC):
        """

        Parameters
        ----------
        XC

        Returns
        -------

        """
        k = self.minmax_threshold[0] + 1  # k         = self.threshold+1;
        neigh = NearestNeighbors(n_neighbors=k, n_jobs=-1)
        neigh.fit(XC)
        dist_, ind = neigh.kneighbors(XC)
        nPt_distance = [dist_[i][k - 1] for i in range(len(dist_))]
        sigma_min = np.quantile(nPt_distance, 0.1)

        k = self.minmax_threshold[1] + 1  # k         = self.threshold+1;
        neigh = NearestNeighbors(n_neighbors=k, n_jobs=-1)
        neigh.fit(XC)
        dist_, ind = neigh.kneighbors(XC)
        nPt_distance = [dist_[i][k - 1] for i in range(len(dist_))]
        sigma_max = np.quantile(nPt_distance, 0.9)

        minmax_sigma = [sigma_min, sigma_max]

        print(
            "Boundaries for sigma    : "
            + str(minmax_sigma[0])
            + " , "
            + str(minmax_sigma[1])
        )

        return minmax_sigma

    def __determine_sigma_boundaries(self, XC):

        """
        OLDER VERSION, THE NEW ONE IS COMMENTED

        Parameters
        ----------
        XC

        Returns
        -------

        """

        k = self.threshold + 1
        # initialize model
        neigh = NearestNeighbors(n_neighbors=k, n_jobs=-1)
        # train for getting nearest neighbour
        neigh.fit(XC)
        dist_, ind = neigh.kneighbors(XC)
        # We have no use of indices here
        # dist is a 2 dimensional array of shape (10000, 9) in which each row is list of length 9. This row contain distances to all 9 nearest points. But we need distance to only 9th nearest point. So
        nPt_distance = [dist_[i][k - 1] for i in range(len(dist_))]

        # CD_sorted = np.sort(dist.squareform(dist.pdist(XC)),axis=1);
        sigma_min = np.quantile(nPt_distance, 0.1)
        sigma_max = np.quantile(nPt_distance, 0.9)
        minmax_sigma = [sigma_min, sigma_max]

        print(
            "Boundaries for sigma    : "
            + str(minmax_sigma[0])
            + " , "
            + str(minmax_sigma[1])
        )

        return minmax_sigma

    def __getLogDistribution(self, min_x, max_x, n):

        """
        Return an array with the log-distribution of 'n' points between
        `min_x` and `max_x`.
        """
        min_log = np.log(min_x)
        max_log = np.log(max_x)

        log_vec = np.linspace(min_log, max_log, n)

        vec = np.exp(log_vec)
        vec = np.unique(vec)  # why is this needed?

        return vec

    def __phaseSpace(self, XC, params):
        """

        Given a set of point to be clustered,
        compute the cluster for each combination of parameters.
        This returns the phase-space,
        which is the set of labels associated with each parameter combination.
        Also computational times are included.

        Parameters
        ----------
        XC:
            data-points
        params:
            list of parameters to compute the phasespace
        Returns
        -------
        ps: pd.Dataframe
        the phase space

        """

        labels_all = []
        times = []

        t1_all = time.time()

        printProgressBar(
            0,
            len(params),
            prefix="Clustering progress:",
            suffix="Complete",
            length=50,
        )

        for index, param in params.iterrows():
            t1 = time.time()
            labels_ = self.ComputeClusters(
                param["sigma"], param["threshold"], XC
            )
            t2 = time.time()
            labels_all.append(labels_)
            times.append(t2 - t1)
            printProgressBar(
                index + 1,
                len(params),
                prefix="Progress:",
                suffix="Complete",
                length=50,
            )
            # print("Computing time for sigma = "+str(np.round(param['sigma'],2))+" and minPts ="+ str(param['threshold'])+" : " + str(np.round(t2-t1,2)) );

        print(
            "Computing clusters : "
            + str(np.round(time.time() - t1_all, 2))
            + " seconds"
        )
        ps = params
        ps["labels"] = labels_all
        ps["time"] = times

        return ps

    def __phaseSpacePostProcess(
            self, XC, PS, skipSimilarityScore: bool = False
    ):

        """
        Postprocess the phasespace, computing the similarities between the clusters.

        **steps**:

        1. Initialize Cluster information
        2. Compute similarity scores (optional)

        Parameters
        ----------
        XC:
            The points to be clustered.
        PS: pd.DataFrame
            The phase-space, i.e., the set of labels for each parameter configuration.
        skipSimilarityScore: bool, default=False
            if 'True', avoid computing the similarities.

        Returns
        -------
        PS: pd.DataFrame
            the phasespace with the postporcessing information.

        """

        print("Postprocessing..")

        n = len(PS)

        no_clusters = np.zeros(shape=(n,), dtype=int)
        similarityScore = np.zeros(shape=(n,))
        # no_locs                  = np.zeros(shape=(n,),dtype=int)
        times = np.zeros(shape=(n,))

        # similarityScoreMatrix    = np.zeros(shape=(n,n))

        # ***********************************************
        # Preprocess: Initialize Cluster information
        # ***********************************************
        clusterInfo = getClusterSizesAll(XC, PS, self.n_dim)
        cli_index = clusterInfo["index"]
        cli_similarityScore = np.zeros(
            [
                len(cli_index),
            ],
            dtype=int,
        )

        for i, ps in PS.iterrows():
            no_clusters[i] = int(np.max(ps["labels"]) + 1)
        PS["no_clusters"] = no_clusters

        # ***************************
        # Compute similarity scores
        # ***************************
        ###

        if skipSimilarityScore == True:
            for i, ps in PS.iterrows():
                similarityScore[i] = np.nan
        else:
            # ***************************************
            # Preprocess: get centers and radii
            # ***************************************

            if self.similarity_score_computation == "total":
                cli_similarityScore, similarityScore = getSimilarityScore(XC, PS, clusterInfo, self.n_dim)
            if self.similarity_score_computation == "threshold":
                (cli_similarityScore, similarityScore) = getSimilarityScoreByThreshold(XC, PS, clusterInfo, self.n_dim)

        clusterInfo["similarityScore"] = cli_similarityScore
        PS["similarityScore"] = similarityScore

        self.clusterInfo = clusterInfo

        return PS

    def __get_consensus_clustering_1d(self, PS, XC):
        """
        Find the best clustering parameters, defined as the one with the highest `similarityScore`.

        This is supposed to be used when optimizing only one parameter (`sigma` or `threshold`).

        Parameters
        ----------
        PS: pd.DataFrame
            the phasepace
        XC:
            The points (#todo: not used, remove)
        Returns
        -------
        labels, selected_parameters:
            the labels of the chosen  configuration and
            the parameters of the chosen configuration
        """
        # take the similarity scores
        similarity = np.asarray(PS["similarityScore"])
        # find the maximum
        max_score = np.max(similarity)
        # take its index
        idx = np.argwhere(similarity == max_score)[-1][0]
        # store the sigma
        sigma_selected = PS.loc[idx, "sigma"]
        # store the threshold
        threshold_selected = PS.loc[idx, "threshold"]
        # store the labels
        labels = PS.loc[idx, "labels"]

        # store the selected parameters
        selected_parameters = {
            "sigma": sigma_selected,
            "threshold": threshold_selected,
        }

        print(
            "Selected threshold , sigma : "
            + str(threshold_selected)
            + " , "
            + str(sigma_selected)
        )
        return labels, selected_parameters

    def __get_consensus_clustering(self, PS, XC, decay=None):
        """
        Find the best parameter configuration fot the 2-d setting, i.e., when optimizing both 'sigma' and 'threshold'.

        **steps**:

        1. Get optimal sigma for each theta
        2. Normalize the line of optima
        3. choose the theta for which the similarity score is < decay

        Parameters
        ----------
        PS: pd.DataFrame
            the phasepace
        XC:
            The points (#todo: not used, remove)
        decay:
            The threshold value for selecting theta among the optima. Must be in (0,1).
        Returns
        -------
        labels, selected_parameters:
            the labels of the chosen  configuration and
            the parameters of the chosen configuration
        """
        if decay is None:
            decay = self.decay
        # 1. Get optimal sigma for each theta
        df_opt_th = GetLineOfOptima(PS, "threshold", "similarityScore")

        line_of_optima = df_opt_th["idx"]
        line_of_optima_sim = np.array(df_opt_th["similarityScore"])
        self.df_opt_th = df_opt_th
        self.line_of_optima = line_of_optima
        # 2. Normalize the diagonal
        opt_normalized = (line_of_optima_sim - line_of_optima_sim.min()) / (
                line_of_optima_sim.max() - line_of_optima_sim.min()
        )

        # 3. choose the theta for which the similarity score is < decay
        ind = np.where(opt_normalized < decay)[0][0]
        optim = PS.iloc[line_of_optima[ind]]

        labels = optim["labels"]
        selected_parameters = {
            "sigma": optim["sigma"],
            "threshold": optim["threshold"],
        }

        print(
            "Selected threshold , sigma : "
            + str(optim["threshold"])
            + " , "
            + str(optim["sigma"])
        )
        return labels, selected_parameters

    def plotPhaseSpace(self, ax=None):

        sigmas = np.unique(self.phasespace["sigma"])
        thresholds = np.unique(self.phasespace["threshold"])

        sigma_opt = self.selected_parameters["sigma"]
        threshold_opt = self.selected_parameters["threshold"]

        sigma_opt_idx = np.where(sigmas == sigma_opt)[0][0]
        threshold_opt_idx = np.where(thresholds == threshold_opt)[0][0]

        index_opt = np.where(sigmas == sigma_opt)[0][0]

        # Compute similarity Matrix
        similarity = []
        for i, row in self.phasespace.iterrows():
            similarity.append(row["similarityScore"])
        sim_matr = np.round(
            np.flipud(np.array(similarity).reshape(15, -1).T), 2
        )

        # Compute line of optima (max_list) for it
        max_list = []
        for t, s in enumerate(sim_matr.argmax(1)):
            max_list.append(s)
        max_list = max_list[::-1]

        # phase space

        if ax is None:
            fig, ax = plt.subplots()

        sns.heatmap(
            sim_matr,
            xticklabels=np.round(sigmas, 2),
            yticklabels=np.flipud(np.round(thresholds)),
            ax=ax,
            cbar=False,
            cmap="Reds",
        )
        from matplotlib.patches import Rectangle

        for i, j in enumerate(max_list[::-1]):
            ax.add_patch(
                Rectangle((j, i), 1, 1, fill=False, edgecolor="blue", lw=3)
            )
        ax.add_patch(
            Rectangle(
                (sigma_opt_idx, len(thresholds) - 1 - threshold_opt_idx),
                1,
                1,
                fill=False,
                edgecolor="green",
                lw=3,
            )
        )
        # sns.lineplot(ax=ax,data=FD.data,x='sigma',y='similarityScore');
        # sns.scatterplot(ax=ax,data=FD.data,x='sigma',y='similarityScore');
        # ax.axvline(sigma_min,c='r');
        # ax.axvline(sigma_max,c='r');
        # ax.axvline(sigma_opt,c='g');
        # #plt.annotate('Selected value', (sigma_opt,FD.phasespace['similarityScore'][index_opt]))
        # trans = ax.get_xaxis_transform()
        # plt.text(sigma_opt, .5, 'Selected value', transform=trans,rotation=90)
        ax.set_xlabel("eps (nm)")
        ax.set_ylabel("minPts")

        plt.show()



# -*- coding: utf-8 -*-

import time

import numpy as np
import pandas as pd
from .AuxiliaryFunctions import printProgressBar

def GetSimilarityScorePerCluster(XC, phasespace, df_clusters=[]):
    if len(df_clusters) == 0:
        df_clusters = getClusterSizesAll(XC, phasespace)

    cli_similarityScore, similarityScore = getSimilarityScore(
        XC, phasespace, df_clusters
    )

    df_clusters["similarityScore"] = cli_similarityScore
    phasespace["similarityScore"] = similarityScore

    return (
        phasespace,
        df_clusters,
    )  # cli_similarityScore,similarityScore; #phaseSpace,df_clusters;


def getSimilarityScoreByThreshold(XC, PS, clusterInfo, n_dim):
    #
    t1 = time.time()
    # Get the info from clusterInfo
    cli_index = clusterInfo["index"]

    cli_similarityScore = np.zeros(
        [
            len(cli_index),
        ],
        dtype=int,
    )
    n = len(PS)
    similarityScoreMatrix = np.zeros(shape=(n, n))
    similarityScore = np.zeros(shape=(n,))
    ps_index = PS.index
    # 1 define a list of the unique threshold values
    threshold_list = np.unique(PS["threshold"])

    # centers,radii = __computeCenters_Radii(XC,PS);

    # Compute the similarities
    progress_i = 0
    printProgressBar(
        progress_i,
        len(PS),
        prefix="Postprocessing progress:",
        suffix="Complete",
        length=50,
    )
    for th in threshold_list:
        PS_th = PS[PS["threshold"] == th]

        for i, ps in PS_th.iterrows():
            mark_i = cli_index == i
            for j in PS_th.index[PS_th.index <= i]:  # np.arange(i+1):
                mark_j = cli_index == j
                if not (i == j):
                    s_ij = getSimilarityScore_ij(i, j, PS, clusterInfo)

                    if (type(s_ij) == bool) and (s_ij == False):
                        continue

                    score = np.sum(s_ij[0])

                    cli_similarityScore[mark_i] += s_ij[0]
                    cli_similarityScore[mark_j] += s_ij[1]

                    similarityScoreMatrix[
                        ps_index == j, ps_index == i
                    ] = score  # /Normalize here?  eg  /int(np.max(PS.loc[j,"labels"]) + 1)
                    similarityScoreMatrix[
                        ps_index == i, ps_index == j
                    ] = score  # /Normalize here?  eg  /int(np.max(PS.loc[i,"labels"]) + 1)
                # else:
                # we do not include similarity with self

            progress_i += 1
            printProgressBar(
                progress_i,
                len(PS),
                prefix="Progress:",
                suffix="Complete",
                length=50,
            )

    print(
        "Computing similarity scores: "
        + str(np.round(time.time() - t1, 2))
        + " seconds"
    )

    # ***************************
    # Collect data
    # ***************************
    for i, ps in PS.iterrows():
        similarityScore[ps_index == i] = np.sum(
            similarityScoreMatrix[ps_index == i, :]
        )

    return cli_similarityScore, similarityScore



def getSimilarityScore(XC, PS, clusterInfo, n_dim):
    """

    Compute similarity score for the whole phase-space.

    **steps**:
    1. Compute the similarities
    2. Collect data

    Parameters
    ----------
    XC:
        The data points (#TODO: not used, remove)
    PS:
        The phase space information

    clusterInfo: pd.DataFrame
        The information about the dataframe, as produced by `getClusterSizesAll`.

    Returns
    -------

    """

    t1 = time.time()
    # Get the info from clusterInfo
    cli_index = clusterInfo["index"]

    cli_similarityScore = np.zeros(
        [
            len(cli_index),
        ],
        dtype=int,
    )
    n = len(PS)
    similarityScoreMatrix = np.zeros(shape=(n, n))
    similarityScore = np.zeros(shape=(n,))
    ps_index = PS.index

    # centers,radii = __computeCenters_Radii(XC,PS);

    # Compute the similarities
    progress_i = 0
    printProgressBar(
        progress_i,
        len(PS),
        prefix="Postprocessing progress:",
        suffix="Complete",
        length=50,
    )
    for i, ps in PS.iterrows():
        mark_i = cli_index == i
        for j in PS.index[PS.index <= i]:  # np.arange(i+1):
            mark_j = cli_index == j

            if not (i == j):
                s_ij = getSimilarityScore_ij(i, j, PS, clusterInfo)

                if (type(s_ij) == bool) and (s_ij == False):
                    continue

                score = np.sum(s_ij[0])

                cli_similarityScore[mark_i] += s_ij[0]
                cli_similarityScore[mark_j] += s_ij[1]

                similarityScoreMatrix[
                    ps_index == j, ps_index == i
                ] = score  # /Normalize here?  eg  /int(np.max(PS.loc[j,"labels"]) + 1)
                similarityScoreMatrix[
                    ps_index == i, ps_index == j
                ] = score  # /Normalize here?  eg  /int(np.max(PS.loc[i,"labels"]) + 1)
            # else:
            # we do not include similarity with self

        progress_i += 1
        printProgressBar(
            progress_i,
            len(PS),
            prefix="Progress:",
            suffix="Complete",
            length=50,
        )
    print(
        "Computing similarity scores: "
        + str(np.round(time.time() - t1, 2))
        + " seconds"
    )

    # ***************************
    # Collect data
    # ***************************
    for i, ps in PS.iterrows():
        similarityScore[ps_index == i] = np.sum(
            similarityScoreMatrix[ps_index == i, :]
        )

    return cli_similarityScore, similarityScore


def getClusterSizesAll(XC, PS, n_dim:int):
    """
    Return a pd.DataFrame with the info about the clustering.
    To do so, compute centers and radii for the subsets in each cluster.
    The information is given by:

    * labels
    * clusterSize
    * threshold
    * sigma
    * index
    * centers
    * radii

    Parameters
    ----------
    XC:
        The points
    PS:
        The phase-space
    n_dim:
        problem dimensionality

    Returns
    -------
    A pd.DataFrame with the info about the clustering.
    """

    cl_sizes = np.array([], dtype=int)
    thresholds = np.array([], dtype=int)
    sigmas = np.array([])
    labels = np.array([], dtype=int)
    index = np.array([], dtype=int)
    centers = np.zeros((0, n_dim))
    radii = np.array([])

    for idx, df1_row in PS.iterrows():

        labels_i = df1_row["labels"]
        l_ = np.arange(np.max(labels_i) + 1)

        if l_.shape[0] == 0:
            continue

        labels = np.concatenate((labels, l_))
        cl_sizes = np.concatenate(
            (cl_sizes, np.asarray([np.sum(labels_i == l) for l in l_]))
        )
        thresholds = np.concatenate(
            (thresholds, df1_row["threshold"] * np.ones_like(l_))
        )
        sigmas = np.concatenate(
            (
                sigmas,
                df1_row["sigma"]
                * np.ones(
                    [
                        len(l_),
                    ]
                ),
            )
        )
        index = np.concatenate((index, idx * np.ones_like(l_)))

        centers_new, radii_new = __computeCenters_Radii_rowPS(XC, df1_row, n_dim)
        centers = np.concatenate((centers, centers_new))
        radii = np.concatenate((radii, radii_new))

    df_clusterSizes = pd.DataFrame()
    df_clusterSizes["labels"] = labels
    df_clusterSizes["clusterSize"] = cl_sizes
    df_clusterSizes["threshold"] = thresholds
    df_clusterSizes["sigma"] = sigmas
    df_clusterSizes["index"] = index
    df_clusterSizes["center"] = list(centers)
    df_clusterSizes["radius"] = radii

    return df_clusterSizes


def getSimilarityScore_ij(i, j, PS, clusterInfo):
    """
    Compute the similarity score between two subsets.

    Parameters
    ----------
    i: int
        first  index
    j: int
        second  index
    PS: pd.DataFrame
        the phase space
    clusterInfo

    Returns
    -------
    [s_i, s_j] : list
        The scores for the i and j component.

    """
    labels_1, labels_2 = PS.loc[i, "labels"], PS.loc[j, "labels"]

    #    clusterInfo_1 = clusterInfo.loc[clusterInfo['index']==i,:];
    #    clusterInfo_2 = clusterInfo.loc[clusterInfo['index']==j,:];
    centers_1 = clusterInfo.loc[clusterInfo["index"] == i, "center"].values
    centers_2 = clusterInfo.loc[clusterInfo["index"] == j, "center"].values

    radii_1 = clusterInfo.loc[clusterInfo["index"] == i, "radius"].values
    radii_2 = clusterInfo.loc[clusterInfo["index"] == j, "radius"].values

    # return zero score if no clusters selected or
    # if one cluster is selected which covers most points
    if (np.max(labels_1) == -1) or (
        np.max(labels_2) == -1
    ):  # cluster are noise
        return False
    elif (np.max(labels_1) == 0) and (
        np.sum(labels_1 == 0) / len(labels_1) > 0.5
    ):
        return False
    elif (np.max(labels_2) == 0) and (
        np.sum(labels_2 == 0) / len(labels_2) > 0.5
    ):
        return False

    # ******************************
    similarityMatrix = np.zeros(
        (np.max(labels_1) + 1, np.max(labels_2) + 1), dtype=int
    )
    similarityMatrix[:] = -1
    # ******************************

    for i1 in np.unique(labels_1):
        if i1 == -1:  # exclude the noise cluster
            continue
        for i2 in np.unique(labels_2):
            if i2 == -1:  # exclude  the noise cluster
                continue

            if similarityMatrix[i1, i2] == 0:
                continue

            if __No_OverlapClusters_Distance(
                i1, i2, centers_1, centers_2, radii_1, radii_2
            ):
                similarityMatrix[i1, i2] = 0
                continue

            if __OverlapClusters_NumberOfLocs(i1, i2, labels_1, labels_2):
                similarityMatrix[i1, :] = 0
                similarityMatrix[:, i2] = 0
                similarityMatrix[i1, i2] = 1
                break
            else:
                similarityMatrix[i1, i2] = 0

    similarityMatrix[similarityMatrix == -1] = 0

    s_i = np.sum(similarityMatrix, axis=1)
    s_j = np.sum(similarityMatrix, axis=0)
    return [s_i, s_j]


def __No_OverlapClusters_Distance(
    i1, i2, centers_1, centers_2, radii_1, radii_2
):
    """

    Return true if the two subsets **do not** overlap.
    This is done by verifying that the distance between their centers is larger than the sum of their radii.

    Parameters
    ----------
    i1: int
        index of the first subset
    i2: int
        index of the second subset
    centers_1:
        list of coordinates of the centers of the first cluster
    centers_2:
        list of coordinates of the centers of the second cluster
    radii_1:
        list of radii of the first cluster
    radii_2:
        list of radii of the second cluster

    Returns
    -------
    bool
    Return True if subset `i1` and subset `i2` do not overlap.
    """
    c1, c2 = centers_1[i1], centers_2[i2]
    r1, r2 = radii_1[i1], radii_2[i2]
    return np.linalg.norm(c2 - c1) > r1 + r2


def __OverlapClusters_NumberOfLocs(i1, i2, labels_1, labels_2):
    """

    Given two labels 'i1' and 'i2' (i.e., two clusters) check if the number of overlapping points among the two is larger
    than the number of non-overlapping points.

    Parameters
    ----------
    i1: int
        label of the first subset
    i2: int
        label of the second subset
    labels_1:
        list of all labels of the first cluster
    labels_2:
        list of all labels of the second cluster
    Returns
    -------
    bool
        True if the number of overlapping points is larger than the number of non-overlapping points.
    """
    # compute the number of element of the first cluster with label == `i1`
    no_locs_1 = np.sum(labels_1 == i1)
    # compute the number of element of the second cluster with label == `i2`
    no_locs_2 = np.sum(labels_2 == i2)
    # compute the overlap between the two clusters (i.e., the number of matching points)
    no_locs_overlap = np.sum((labels_1 == i1) * (labels_2 == i2))

    # return true if the majority of points overlaps.
    if (no_locs_overlap / no_locs_1 > 0.5) and (
        no_locs_overlap / no_locs_2 > 0.5
    ):
        return True
    else:
        return False


def __computeCenters_Radii_rowPS(XC, ps, n_dim):
    """
    For each parameter configuration in a single row of the dataset, compute the center and the radius of each cluster.

    * `centers`:
    a list of arrays each of dimension (`no_clusters`,2),
    storing the coordinate of the center of each cluster
    Note that `no_clusters` varies for each element of the list.

    * 'radii':
    a list of arrays of size ('no_clusters').
    Contains the radius of each cluster, defined as the distance of the farthest point from the center.
    Note that `no_clusters` varies for each element of the list.


        Parameters
        ----------
        XC:
            the datapoints.
        ps:
            a row of the phaspace.
        n_dim:
            problem dimensionality

        Returns
        -------
        centers, radii
    """

    no_clusters = np.max(ps["labels"]) + 1

    centers_i = np.zeros((no_clusters, n_dim))
    radii_i = np.zeros((no_clusters,))

    # go through all clusters:
    for icl in np.unique(ps["labels"]):
        if icl == -1:
            continue
        XCm = XC[(ps["labels"] == icl)]

        c = np.median(XCm, axis=0)
        centers_i[icl, :] = c
        radii_i[icl] = np.max(np.linalg.norm(XCm - c, axis=1))

    return centers_i, radii_i


def __computeCenters_Radii(XC, PS):
    """
    #TODO: this is not used!

    For each parameter configuration, compute the center and the radius of each cluster.

    * `centers` :

    a list of arrays each of dimension (`no_clusters`,2),
    storing the coordinate of the center of each cluster
    Note that `no_clusters` varies for each element of the list.

    * `radii` :

    a list of arrays of size ('no_clusters').
    Contains the radius of each cluster, defined as the distance of the farthest point from the center.
    Note that `no_clusters` varies for each element of the list.


        Parameters
        ----------
        XC:
            the datapoints
        PS:
            the phase space

        Returns
        -------
        centers, radii
    """

    centers = []
    radii = []

    # loop through the data
    for i, ps in PS.iterrows():
        # compute the number of labels for that particular configuration of parameters.
        no_clusters = np.max(ps["labels"]) + 1
        # initialize the centers and the radii for the configuration
        centers_i = np.zeros((no_clusters, 2))
        radii_i = np.zeros((no_clusters,))

        # go through all clusters:
        for icl in np.unique(ps["labels"]):
            if icl == -1:  # ignore noise
                continue
            # take all the point  with label icl
            XCm = XC[(ps["labels"] == icl)]
            # compute the median
            c = np.median(XCm, axis=0)
            # store the center position in the icl position
            centers_i[icl, :] = c
            # store the radius of the cluster, computed as the distance of the farthest point from the center.
            radii_i[icl] = np.max(np.linalg.norm(XCm - c, axis=1))

        centers.append(centers_i)
        radii.append(radii_i)

    return centers, radii
